(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
    typeof define === 'function' && define.amd ? define(['exports'], factory) :
    (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.Rasti = {}));
})(this, (function (exports) { 'use strict';

    /**
     * `Emitter` is a class that provides an easy way to implement the observer pattern 
     * in your applications.  
     * It can be extended to create new classes that have the ability to emit and bind custom named events.   
     * Emitter is used by `Model` and `View` classes, which inherit from it to implement 
     * event-driven functionality.
     *
     * @module
     * @example
     * import { Emitter } from 'rasti';
     * // Custom cart
     * class ShoppingCart extends Emitter {
     *     constructor() {
     *         super();
     *         this.items = [];
     *     }
     *
     *     addItem(item) {
     *         this.items.push(item);
     *         // Emit a custom event called `itemAdded`.
     *         // Pass the added item as an argument to the event listener.
     *         this.emit('itemAdded', item);
     *     }
     * }
     * // Create an instance of ShoppingCart and Logger
     * const cart = new ShoppingCart();
     * // Listen to the `itemAdded` event and log the added item using the logger.
     * cart.on('itemAdded', (item) => {
     *     console.log(`Item added to cart: ${item.name} - Price: $${item.price}`);
     * });
     * // Simulate adding items to the cart
     * const item1 = { name : 'Smartphone', price : 1000 };
     * const item2 = { name : 'Headphones', price : 150 };
     *
     * cart.addItem(item1); // Output: "Item added to cart: Smartphone - Price: $1000"
     * cart.addItem(item2); // Output: "Item added to cart: Headphones - Price: $150"
     */
    class Emitter {
        /**
         * Adds event listener.
         * @param {string} type Type of the event (e.g. `change`).
         * @param {function} listener Callback function to be called when the event is emitted.
         * @example
         * // Re render when model changes.
         * this.model.on('change', this.render.bind(this));
         */
        on(type, listener) {
            // Validate listener.
            if (typeof listener !== 'function') {
                throw new TypeError('Listener must be a function');
            }
            // Create listeners object if it doesn't exist.
            if (!this.listeners) this.listeners = {};
            if (!this.listeners[type]) this.listeners[type] = [];
            // Add listener.
            this.listeners[type].push(listener);
            // Return a function to remove the listener.
            return () => this.off(type, listener);
        }

        /**
         * Adds event listener that executes once.
         * @param {string} type Type of the event (e.g. `change`).
         * @param {function} listener Callback function to be called when the event is emitted.
         * @example
         * // Log a message once when model changes.
         * this.model.once('change', () => console.log('This will happen once'));
         */
        once(type, listener) {
            // If listener is a function, wrap it to remove it after it is called.
            if (typeof listener === 'function') {
                const self = this;
                const originalListener = listener;

                listener = function(...args) {
                    originalListener(...args);
                    self.off(type, listener);
                };
            }
            // Add listener.
            return this.on(type, listener);
        }

        /**
         * Removes event listeners.
         * @param {string} [type] Type of the event (e.g. `change`). If is not provided, it removes all listeners.
         * @param {function} [listener] Callback function to be called when the event is emitted. If listener is not provided, it removes all listeners for specified type.
         * @example
         * // Stop listening to changes.
         * this.model.off('change');
         */
        off(type, listener) {
            // No listeners.
            if (!this.listeners) return;
            // No type provided, remove all listeners.
            if (!type) {
                delete this.listeners;
                return;
            }
            // No listeners for specified type.
            if (!this.listeners[type]) return;
            // No listener provided, remove all listeners for specified type.
            if (!listener) {
                delete this.listeners[type];
            } else {
                // Remove specific listener.
                this.listeners[type] = this.listeners[type].filter(fn => fn !== listener);
                if (!this.listeners[type].length) delete this.listeners[type];
            }
            // Remove listeners object if it's empty.
            if (!Object.keys(this.listeners).length) delete this.listeners;
        }

        /**
         * Emits event of specified type. Listeners will receive specified arguments.
         * @param {string} type Type of the event (e.g. `change`).
         * @param {any} [...args] Arguments to be passed to listener.
         * @example
         * // Emit validation error event.
         * this.emit('invalid');
         */
        emit(type, ...args) {
            // No listeners.
            if (!this.listeners || !this.listeners[type]) return;
            // Call listeners. Use `slice` to make a copy and prevent errors when 
            // removing listeners inside a listener.
            this.listeners[type]
                .slice()
                .forEach(function(fn) {
                    fn(...args);
                });
        }
    }

    /*
     * Evaluate the expression. If it's a function, call it with the provided context and return the result.
     * Otherwise, return the expression as is.
     * @param {any} expression Expression to be evaluated.
     * @param {any} context Context to call the expression.
     * @param {...any} args Arguments to pass to the expression.
     * @return {any} The result of the expression.
     */
    var getResult = (expression, context, ...args) =>
        typeof expression !== 'function' ? expression :
            expression.apply(context, args);

    /**
     * - Orchestrates data and business logic.
     * - Emits events when data changes.
     * 
     * A `Model` manages an internal table of data attributes and triggers change events when any of its data is modified.  
     * Models may handle syncing data with a persistence layer. To design your models, create atomic, reusable objects 
     * that contain all the necessary functions for manipulating their specific data.  
     * Models should be easily passed throughout your app and used anywhere the corresponding data is needed.  
     * Rasti models store their attributes in `this.attributes`, which is extended from `this.defaults` and the 
     * constructor `attributes` parameter. For every attribute, a getter is generated to retrieve the model property 
     * from `this.attributes`, and a setter is created to set the model property in `this.attributes` and emit `change` 
     * and `change:attribute` events.
     * @module
     * @extends Rasti.Emitter
     * @param {object} attributes Object containing model attributes to extend `this.attributes`. Getters and setters are generated for `this.attributes`, in order to emit `change` events.
     * @property {object|function} defaults Object containing default attributes for the model. It will extend `this.attributes`. If a function is passed, it will be called to get the defaults. It will be bound to the model instance.
     * @property {object} previous Object containing previous attributes when a change occurs.
     * @example
     * import { Model } from 'rasti';
     * // Product model
     * class ProductModel extends Model {
     *     preinitialize() {
     *         // The Product model has `name` and `price` default attributes.
     *         // `defaults` will extend `this.attributes`.
     *         // Getters and setters are generated for `this.attributes`,
     *         // in order to emit `change` events.
     *         this.defaults = {
     *             name: '',
     *             price: 0
     *         };
     *     }
     *
     *     setDiscount(discountPercentage) {
     *         // Apply a discount to the price property.
     *         // This will call a setter that will update `price` in `this.attributes`,
     *         // and emit `change` and `change:price` events.
     *         const discount = this.price * (discountPercentage / 100);
     *         this.price -= discount;
     *     }
     * }
     * // Create a product instance with a name and price.
     * const product = new ProductModel({ name: 'Smartphone', price: 1000 });
     * // Listen to the `change:price` event.
     * product.on('change:price', () => console.log('New Price:', product.price));
     * // Apply a 10% discount to the product.
     * product.setDiscount(10); // Output: "New Price: 900"
     */
    class Model extends Emitter {
        constructor(attributes = {}) {
            super();
            // Call preinitialize.
            this.preinitialize.apply(this, arguments);
            // Get defaults. If `this.defaults` is a function, call it.
            const defaults = getResult(this.defaults, this) || {};
            // Set attributes object with defaults and passed attributes.
            this.attributes = Object.assign({}, defaults, attributes);
            // Object to store previous attributes when a change occurs.
            this.previous = {};
            // Generate getters/setters for every attribute.
            Object.keys(this.attributes).forEach(this.defineAttribute.bind(this));
        }

        /**
         * If you define a preinitialize method, it will be invoked when the Model is first created, before any instantiation logic is run for the Model.
         * @param {object} attributes Object containing model attributes to extend `this.attributes`.
         */
        preinitialize() {}

        /**
         * Generate getter/setter for the given key. In order to emit `change` events.
         * This method is called internally by the constructor
         * for `this.attributes`.
         * @param {string} key Attribute key.
         */
        defineAttribute(key) {
            Object.defineProperty(
                this,
                key, {
                    get : () => this.get(key),
                    set : (value) => { this.set(key, value); }
                }
            );
        }

        /**
         * Get an attribute from `this.attributes`.
         * This method is called internally by generated getters.
         * @param {string} key Attribute key.
         * @return {any} The attribute value.
         */
        get(key) {
            return this.attributes[key];
        }

        /**
         * Set an attribute into `this.attributes`.  
         * Emit `change` and `change:attribute` if a value changes.  
         * Could be called in two forms, `this.set('key', value)` and
         * `this.set({ key : value })`.  
         * This method is called internally by generated setters.  
         * The `change` event listener will receive the model instance, an object containing the changed attributes, and the rest of the arguments passed to `set` method.  
         * The `change:attribute` event listener will receive the model instance, the new attribute value, and the rest of the arguments passed to `set` method.
         * @param {string} key Attribute key or object containing keys/values.
         * @param [value] Attribute value.
         * @return {this} This model.
         * @emits change
         * @emits change:attribute
         */
        set(key, value, ...rest) {
            let attrs, args;
            // Handle both `"key", value` and `{key: value}` style arguments.
            if (typeof key === 'object') {
                attrs = key;
                args = [value, ...rest];
            } else {
                attrs = { [key] : value };
                args = rest;
            }
            // Are we in a nested `set` call?
            // Calling a `set` inside a `change:attribute` or `change` event listener
            const changing = this._changing;
            this._changing = true;
            // Store changed attributes.
            const changed = {};
            // Store previous attributes.
            if (!changing) {
                this.previous = Object.assign({}, this.attributes);
            }
            // Set attributes.
            Object.keys(attrs).forEach(key => {
                // Use equality to determine if value changed.
                if (attrs[key] !== this.attributes[key]) {
                    changed[key] = attrs[key];
                    this.attributes[key] = attrs[key];
                }
            });

            const changedKeys = Object.keys(changed);
            // Pending `change` event arguments.
            if (changedKeys.length) this._pending = ['change', this, changed, ...args];
            // Emit `change:attribute` events.
            changedKeys.forEach(key => {
                this.emit(`change:${key}`, this, attrs[key], ...args);
            });
            // Don't emit `change` event until the end of the nested 
            // `set` calls inside `change:attribute` event listeners.
            if (changing) return this;
            // Emit `change` events, that might be nested.
            while (this._pending) {
                const pendingChange = this._pending;
                this._pending = null;
                this.emit.apply(this, pendingChange);
            }
            // Reset flags.
            this._pending = null;
            this._changing = false;

            return this;
        }

        /**
         * Return object representation of the model to be used for JSON serialization.
         * By default returns a copy of `this.attributes`.
         * @return {object} Object representation of the model to be used for JSON serialization.
         */
        toJSON() {
            return Object.assign({}, this.attributes);
        }
    }

    /*
     * These option keys will be extended on the view instance.
     */
    const viewOptions = ['el', 'tag', 'attributes', 'events', 'model', 'template', 'onDestroy'];

    /**
     * - Listens for changes and renders the UI.
     * - Handles user input and interactivity.
     * - Sends captured input to the model.
     *
     * A `View` is an atomic unit of the user interface that can render data from a specific model or multiple models.
     * However, views can also be independent and have no associated data.  
     * Models must be unaware of views. Views, on the other hand, may render model data and listen to the change events 
     * emitted by the models to re-render themselves based on changes.  
     * Each `View` has a root element, `this.el`, which is used for event delegation.  
     * All element lookups are scoped to this element, and any rendering or DOM manipulations should be done inside it. 
     * If `this.el` is not present, an element will be created using `this.tag` (defaulting to `div`) and `this.attributes`.
     * @module
     * @extends Emitter
     * @param {object} options Object containing options. The following keys will be merged into the view instance: `el`, `tag`, `attributes`, `events`, `model`, `template`, `onDestroy`.
     * @property {node|function} el Every view has a root DOM element stored at `this.el`. If not present, it will be created. If `this.el` is a function, it will be called to get the element at `this.ensureElement`, bound to the view instance. See {@link module_view__ensureelement View.ensureElement}.
     * @property {string|function} tag If `this.el` is not present, an element will be created using `this.tag` and `this.attributes`. Default is `div`. If it is a function, it will be called to get the tag, bound to the view instance. See {@link module_view__ensureelement View.ensureElement}.
     * @property {object|function} attributes If `this.el` is not present, an element will be created using `this.tag` and `this.attributes`. If it is a function, it will be called to get the attributes object, bound to the view instance. See {@link module_view__ensureelement View.ensureElement}.
     * @property {object|function} events Object in the format `{'event selector' : 'listener'}`. It will be used to bind delegated event listeners to the root element. If it is a function, it will be called to get the events object, bound to the view instance. See {@link module_view_delegateevents View.delegateEvents}.
     * @property {object} model A model or any object containing data and business logic.
     * @property {function} template A function that returns a string with the view's inner HTML. See {@link module_view__render View.render}. 
     * @property {number} uid Unique identifier for the view instance. This can be used to generate unique IDs for elements within the view. It is automatically generated and should not be set manually.
     * @example
     * import { View } from 'rasti';
     * 
     * class Timer extends View {
     *     constructor(options) {
     *         super(options);
     *         // Create model to store internal state. Set `seconds` attribute to 0.
     *         this.model = new Model({ seconds : 0 });
     *         // Listen to changes in model `seconds` attribute and re-render.
     *         this.model.on('change:seconds', this.render.bind(this));
     *         // Increment model `seconds` attribute every 1000 milliseconds.
     *         this.interval = setInterval(() => this.model.seconds++, 1000);
     *     }
     *
     *     template(model) {
     *         return `Seconds: <span>${model.seconds}</span>`;
     *     }
     * }
     * // Render view and append view's element into the body.
     * document.body.appendChild(new Timer().render().el);
     */
    class View extends Emitter {
        constructor(options = {}) {
            super();
            // Call preinitialize.
            this.preinitialize.apply(this, arguments);
            // Generate unique id.
            // Useful to generate element ids.
            this.uid = `rasti-${++View.uid}`;
            // Store delegated event listeners,
            // so they can be unbound later.
            this.delegatedEventListeners = [];
            // Store child views,
            // so they can be destroyed.
            this.children = [];
            // Mutable array to store handlers to be called on destroy.
            this.destroyQueue = [];
            // Extend "this" with options.
            viewOptions.forEach(key => {
                if (key in options) this[key] = options[key];
            });
            // Ensure that the view has a root element at `this.el`.
            this.ensureElement();
        }

        /**
         * If you define a preinitialize method, it will be invoked when the view is first created, before any instantiation logic is run.
         * @param {object} attrs Object containing model attributes to extend `this.attributes`.
         */
        preinitialize() {}

        /**
         * Returns the first element that matches the selector, 
         * scoped to DOM elements within the current view's root element (`this.el`).
         * @param {string} selector CSS selector.
         * @return {node} Element matching selector within the view's root element (`this.el`).
         */
        $(selector) {
            return this.el.querySelector(selector);
        }

        /**
         * Returns a list of elements that match the selector, 
         * scoped to DOM elements within the current view's root element (`this.el`).
         * @param {string} selector CSS selector.
         * @return {node[]} List of elements matching selector within the view's root element (`this.el`).
         */
        $$(selector) {
            return this.el.querySelectorAll(selector);
        }

        /**
         * Destroy the view.
         * Destroy children views if any, undelegate events, stop listening to events, call `onDestroy` lifecycle method.
         * @param {object} options Options object or any arguments passed to `destroy` method will be passed to `onDestroy` method.
         * @return {Rasti.View} Return `this` for chaining.
         */
        destroy() {
            // Call destroy on children.
            this.destroyChildren();
            // Undelegate `this.el` event listeners
            this.undelegateEvents();
            // Unbind `this` events.
            this.off();
            // Call destroy queue.
            this.destroyQueue.forEach(fn => fn());
            this.destroyQueue = [];
            // Call onDestroy lifecycle method
            this.onDestroy.apply(this, arguments);
            // Return `this` for chaining.
            return this;
        }

        /**
         * `onDestroy` lifecycle method is called after the view is destroyed.
         * Override with your code. Useful to stop listening to model's events.
         * @param {object} options Options object or any arguments passed to `destroy` method.
         */
        onDestroy() {}

        /**
         * Add a view as a child.
         * Children views are stored at `this.children`, and destroyed when the parent is destroyed.
         * Returns the child for chaining.
         * @param {Rasti.View} child
         * @return {Rasti.View}
         */
        addChild(child) {
            this.children.push(child);
            return child;
        }

        /**
         * Call destroy method on children views.
         */
        destroyChildren() {
            this.children.forEach(child => child.destroy());
            this.children = [];
        }

        /**
         * Ensure that the view has a root element at `this.el`.
         * You shouldn't call this method directly. It's called from the constructor.
         * You may override it if you want to use a different logic or to 
         * postpone element creation.
         */ 
        ensureElement() {
            // Element is already present.
            if (this.el) {
                // If "this.el" is a function, call it to get the element.
                this.el = getResult(this.el, this);
            } else {
                // If "this.el" is not present,
                // create a new element according "this.tag"
                // and "this.attributes".
                const tag = getResult(this.tag, this);
                const attrs = getResult(this.attributes, this);
                this.el = this.createElement(tag, attrs);
            }
            // Delegate events on element.
            this.delegateEvents();
        }

        /**
         * Create an element.
         * Called from the constructor if `this.el` is undefined, to ensure
         * the view has a root element.
         * @param {string} tag Tag for the element. Default to `div`
         * @param {object} attributes Attributes for the element.
         * @return {node} The created element.
         */
        createElement(tag = 'div', attributes = {}) {
            // Create DOM element.
            let el = document.createElement(tag);
            // Add element attributes.
            Object.keys(attributes)
                .forEach(key => el.setAttribute(key, attributes[key]));

            return el;
        }

        /**
         * Remove `this.el` from the DOM.
         * @return {Rasti.View} Return `this` for chaining.
         */
        removeElement() {
            this.el.parentNode.removeChild(this.el);
            // Return `this` for chaining.
            return this;
        }

        /**
         * Provide declarative listeners for DOM events within a view. If an events object is not provided, 
         * it defaults to using `this.events`. If `this.events` is a function, it will be called to get the events object.
         * 
         * The events object should follow the format `{'event selector': 'listener'}`:
         * - `event`: The type of event (e.g., 'click').
         * - `selector`: A CSS selector to match the event target. If omitted, the event is bound to the root element.
         * - `listener`: A function or a string representing a method name on the view. The method will be called with `this` bound to the view instance.
         * 
         * By default, `delegateEvents` is called within the View's constructor. If you have a simple events object, 
         * all of your DOM events will be connected automatically, and you will not need to call this function manually.
         * 
         * All attached listeners are bound to the view, ensuring that `this` refers to the view object when the listeners are invoked.
         * When `delegateEvents` is called again, possibly with a different events object, all previous listeners are removed and delegated afresh.
         * 
         * The listeners will be invoked with the event and the view as arguments.
         * 
         * @param {object} [events] Object in the format `{'event selector' : 'listener'}`. Used to bind delegated event listeners to the root element.
         * @return {Rasti.View} Returns `this` for chaining.
         * @example
         * // Using a function.
         * class Modal extends View {
         *     events() {
         *         return {
         *             'click button.ok': 'onClickOkButton',
         *             'click button.cancel': function() {}
         *         };
         *     }
         * }
         * 
         * // Using an object.
         * Modal.prototype.events = {
         *     'click button.ok' : 'onClickOkButton',
         *     'click button.cancel' : function() {}
         * };
         */
        delegateEvents(events) {
            if (!events) events = getResult(this.events, this);
            if (!events) return this;

            if (this.delegatedEventListeners.length) this.undelegateEvents();

            // Store events by type i.e.: "click", "submit", etc.
            let eventTypes = {};

            Object.keys(events).forEach(key => {
                const keyParts = key.split(' ');
                const type = keyParts.shift();
                const selector = keyParts.join(' ');

                let listener = events[key];
                // Listener may be a string representing a method name on the view,
                // or a function.
                listener = (
                    typeof listener === 'string' ?
                        this[listener] :
                        listener
                ).bind(this);

                if (!eventTypes[type]) eventTypes[type] = [];

                eventTypes[type].push({ selector, listener });
            });

            Object.keys(eventTypes).forEach(type => {
                // Listener for the type of event.
                const typeListener = (event) => {
                    // Iterate and run every individual listener if the selector matches.
                    eventTypes[type].forEach(({ selector, listener }) => {
                        if (!selector || event.target.closest(selector)) listener(event, this);
                    });
                };

                this.delegatedEventListeners.push({ type, listener : typeListener });
                this.el.addEventListener(type, typeListener);
            });
            // Return `this` for chaining.
            return this;
        }

        /**
         * Removes all of the view's delegated events. 
         * Useful if you want to disable or remove a view from the DOM temporarily. 
         * Called automatically when the view is destroyed and when `delegateEvents` is called again.
         * @return {Rasti.View} Return `this` for chaining.
         */
        undelegateEvents() {
            this.delegatedEventListeners.forEach(({ type, listener }) => {
                this.el.removeEventListener(type, listener);
            });

            this.delegatedEventListeners = [];
            // Return `this` for chaining.
            return this;
        }

        /**
         * Renders the view.  
         * This method should be overridden with custom logic.
         * The only convention is to manipulate the DOM within the scope of `this.el`,
         * and to return `this` for chaining.  
         * If you add any child views, you should call `this.destroyChildren` before re-rendering.  
         * The default implementation updates `this.el`'s innerHTML with the result
         * of calling `this.template`, passing `this.model` as the argument.
         * <br><br> &#9888; **Security Notice:** The default implementation utilizes `innerHTML`, which may introduce Cross-Site Scripting (XSS) risks.  
         * Ensure that any user-generated content is properly sanitized before inserting it into the DOM. 
         * You can use the {@link #module_view_sanitize View.sanitize} static method to escape HTML entities in a string.  
         * For best practices on secure data handling, refer to the 
         * [OWASP's XSS Prevention Cheat Sheet](https://cheatsheetseries.owasp.org/cheatsheets/Cross_Site_Scripting_Prevention_Cheat_Sheet.html).<br><br>
         * @return {Rasti.View} Returns `this` for chaining.
         */
        render() {
            if (this.template) this.el.innerHTML = this.template(this.model);
            // Return `this` for chaining.
            return this;
        }

        /**
         * Escape HTML entities in a string.
         * Use this method to sanitize user-generated content before inserting it into the DOM.
         * Override this method to provide a custom escape function.
         * This method is inherited by {@link #module_component Component} and used to escape template interpolations.
         * @static
         * @param {string} str String to escape.
         * @return {string} Escaped string.
         */
        static sanitize(value) {
            return `${value}`.replace(/[&<>"']/g, match => ({
                '&' : '&amp;',
                '<' : '&lt;',
                '>' : '&gt;',
                '"' : '&quot;',
                '\'' : '&#039;'
            }[match]));
        }
    }

    /**
     * Counter for generating unique IDs for view instances.  
     * This is primarily used to assign unique identifiers to each view instance (`this.uid`), which can be helpful for tasks like 
     * generating element IDs.  
     * {@link #module_component Component}s use `this.uid` to generate data attributes for their elements, to be looked up on hydration.  
     * For server-side rendering, this counter should be reset to `0` on every request to ensure that the generated 
     * unique IDs match those on the client, enabling seamless hydration of components.  
     * @static
     * @type {number}
     * @default 0
     */
    View.uid = 0;

    /*
     * Flatten an array recursively
     * @param {Array} arr Array to flat recursively
     * @return {Array} Flat array
     */
    const deepFlat = (arr) => arr.reduce((acc, val) => {
        if (Array.isArray(val)) acc.push(...deepFlat(val));
        else acc.push(val);
        return acc;
    }, []);

    /**
     * Wrapper class for HTML strings marked as safe.
     * @class SafeHTML
     * @param {string} value The HTML string to be marked as safe.
     * @property {string} value The HTML string.
     * @private
     */
    class SafeHTML {
        constructor(value) {
            this.value = value;
        }

        toString() {
            return this.value;
        }
    }

    /**
     * Same as getResult, but pass context as argument to the expression.
     * Used to evaluate expressions in the context of a component.
     * @param {any} expression The expression to be evaluated.
     * @param {any} context The context to call the expression with.
     * @return {any} The result of the evaluated expression.
     * @private
     */
    const getExpressionResult = (expression, context) => getResult(expression, context, context);

    /**
     * Generate string with placeholders for interpolated expressions.
     * @param strings {array} Array of strings.
     * @param expressions {array} Array of expressions.
     * @return {string} String with placeholders.
     * @private
     */
    const addPlaceholders = (strings, expressions) => 
        strings.reduce((out, string, i) => {
            // Add string part.
            out.push(string);
            // Add expression placeholders.
            if (typeof expressions[i] !== 'undefined') {
                out.push(Component.PLACEHOLDER_EXPRESSION(i));
            }
            return out;
        }, []).join('');

    /**
     * Generate one dimensional array with strings and expressions.
     * @param main {string} The main template containing placeholders.
     * @param expressions {array} Array of expressions to replace placeholders.
     * @return {array} Array containing strings and expressions.
     * @private
     */
    const splitPlaceholders = (main, expressions) => {
        const PH = Component.PLACEHOLDER_EXPRESSION('(\\d+)');
        const regExp = new RegExp(`${PH}`, 'g');
        const out = [];
        let lastIndex = 0;
        let match;
        // Generate one dimensional array with strings and expressions, 
        // so all the components are added as children by the parent component.
        while ((match = regExp.exec(main)) !== null) {
            const before = main.slice(lastIndex, match.index);
            out.push(Component.markAsSafeHTML(before), expressions[match[1]]);
            lastIndex = match.index + match[0].length;
        }
        out.push(Component.markAsSafeHTML(main.slice(lastIndex)));

        return out;
    };

    /**
     * Expand attributes.
     * @param attributes {array} Array of attributes as key, value pairs.
     * @param getExpressionResult {function} Function to render expressions.
     * @return {object}
     * @property {object} all All attributes.
     * @property {object} events Event listeners.
     * @property {object} attributes Attributes.
     * @private
     */
    const expandAttributes = (attributes, getExpressionResult) => {
        const out = attributes.reduce((out, pair) => {
            const attribute = getExpressionResult(pair[0]);
            // Attribute without value. 
            if (pair.length === 1) {
                if (typeof attribute === 'object') {
                    // Expand objects as attributes.
                    out.all = Object.assign(out.all, attribute);
                } else if (typeof attribute === 'string') {
                    // Treat as boolean.
                    out.all[attribute] = true;
                }
            } else {
                // Attribute with value.
                const value = getExpressionResult(pair[1]);
                out.all[attribute] = value;
            }

            return out;
        }, { all : {}, events : {}, attributes : {} });

        Object.keys(out.all).forEach(key => {
            // Check if key is an event listener.
            const match = key.match(/on(([A-Z]{1}[a-z]+)+)/);
            if (match && match[1]) {
                // Add event listener.
                out.events[match[1].toLowerCase()] = out.all[key];
            } else {
                // Add attribute.
                out.attributes[key] = out.all[key];
            }
        });

        return out;
    };

    /**
     * Replace component tags with expressions.
     * `<${Component} />` or `<${Component}></${Component}>` will be replaced 
     * by a function that mounts the component.
     * Returns the template with component tags replaced by expressions placeholders 
     * modifies the expressions array adding the mount functions.
     * @param main {string} The main template.
     * @return {string} The template with components tags replaced by expressions
     * placeholders.
     * @private
     */
    const expandComponents = (main, expressions) => {
        const PH = Component.PLACEHOLDER_EXPRESSION('(\\d+)');
        // Match component tags.
        return main.replace(
            new RegExp(`<(${PH})([^>]*)>([\\s\\S]*?)</(${PH})>|<(${PH})([^>]*)/>`,'g'),
            function() {
                const { tag, attributes, inner, close, raw } = parseMatch(arguments, expressions);
                // No component found.
                if (!(tag.prototype instanceof Component)) return raw;

                let renderChildren;
                // Non void component.
                if (close) {
                    // Close component tag must match open component tag.
                    if (tag !== close) return raw;
                    // Recursively expand inner components.
                    const list = splitPlaceholders(expandComponents(inner, expressions), expressions);
                    // Create renderChildren function.
                    renderChildren = function() {
                        return deepFlat(list.map(item => getExpressionResult(item, this)));
                    };
                }
                // Create mount function.
                const mount = function() {
                    const options = expandAttributes(attributes, value => getExpressionResult(value, this)).all;
                    // Add renderChildren function to options.
                    if (renderChildren) options.renderChildren = renderChildren.bind(this);
                    // Mount component.
                    return tag.mount(options);
                };
                // Add mount function to expression.
                expressions.push(mount);
                // Replace whole string with expression placeholder.
                return Component.PLACEHOLDER_EXPRESSION(expressions.length - 1);
            }
        );
    };

    /**
     * Parse match data to get tag, attributes, inner html and close tag.
     * @param match {array}
     * @return {object}
     * @property {string} tag The tag.
     * @property {string} inner The inner html.
     * @property {string} close The closing tag.
     * @property {array} attributes Array of attributes as key, value pairs.
     * @property {string} raw The whole match.
     * @private
     */
    const parseMatch = (match, expressions) => {
        const PH = Component.PLACEHOLDER_EXPRESSION('(\\d+)');

        const [all, openTag, openIdx, nonVoidAttrs, inner, closeTag, closeIdx,
            selfClosingTag, selfClosingIdx, selfClosingAttrs] = match;

        const data = { raw : all, attributes : [] };

        let attributesStr;
        if (openTag) {
            // Non void element.
            data.tag = typeof openIdx !== 'undefined' ? expressions[openIdx] : openTag;
            data.inner = inner;
            data.close = typeof closeIdx !== 'undefined' ? expressions[closeIdx] : closeTag;
            attributesStr = nonVoidAttrs;
        } else {
            // Self closing element.
            data.tag = typeof selfClosingIdx !== 'undefined' ? expressions[selfClosingIdx] : selfClosingTag;
            attributesStr = selfClosingAttrs;
        }
        // Parse attributes.
        const regExp = new RegExp(`(${PH}|[\\w-]+)(?:=(["']?)(?:${PH}|((?:.?(?!["']?\\s+(?:\\S+)=|\\s*/?[>"']))+.))\\3)?`, 'g');

        let attributeMatch;
        while ((attributeMatch = regExp.exec(attributesStr)) !== null) {
            const [, attribute, attributeIdx,, valueIdx, value] = attributeMatch;

            const attr = typeof attributeIdx !== 'undefined' ? expressions[attributeIdx] : attribute;
            const val = typeof valueIdx !== 'undefined' ? expressions[valueIdx] : value;

            if (typeof val !== 'undefined') {
                data.attributes.push([attr, val]);
            } else {
                data.attributes.push([attr]);
            }
        }

        return data;
    };

    /*
     * HTML tags that are self closing.
     */
    const selfClosingTags = {
        area : true, base : true, br : true, col : true, embed : true, hr : true, 
        img : true, input : true, link : true, meta : true, source : true, track : true, wbr : true
    };

    /*
     * These option keys will be extended on the component instance.
     */
    const componentOptions = ['key', 'state', 'onCreate', 'onChange', 'onRender'];

    /**
     * Components are a special kind of `View` that is designed to be easily composable, 
     * making it simple to add child views and build complex user interfaces.  
     * Unlike views, which are render-agnostic, components have a specific set of rendering 
     * guidelines that allow for a more declarative development style.  
     * Components are defined with the {@link #module_component_create Component.create} static method, which takes a tagged template string or a function that returns another component.
     * @module
     * @extends Rasti.View
     * @param {object} options Object containing options. The following keys will be merged to `this`: model, state, key, onDestroy, onRender, onCreate, onChange.
     * @property {string} key A unique key to identify the component. Used to recycle child components.
     * @property {object} model A `Rasti.Model` or any emitter object containing data and business logic. The component will listen to `change` events and call `onChange` lifecycle method.
     * @property {object} state A `Rasti.Model` or any emitter object containing data and business logic, to be used as internal state. The component will listen to `change` events and call `onChange` lifecycle method.
     * @see {@link #module_component_create Component.create}
     * @example
     * import { Component, Model } from 'rasti';
     * // Create Timer component.
     * const Timer = Component.create`
     *     <div>
     *         Seconds: <span>${({ model }) => model.seconds}</span>
     *     </div>
     * `;
     * // Create model to store seconds.
     * const model = new Model({ seconds: 0 });
     * // Mount timer on body.
     * Timer.mount({ model }, document.body);
     * // Increment `model.seconds` every second.
     * setInterval(() => model.seconds++, 1000);
     */
    class Component extends View {
        constructor(options = {}) {
            super(...arguments);
            // Extend "this" with options.
            componentOptions.forEach(key => {
                if (key in options) this[key] = options[key];
            });
            // Store options by default.
            this.options = options;
            // Bind `partial` method to `this`.
            this.partial = this.partial.bind(this);
            // Call lifecycle method.
            this.onCreate.apply(this, arguments);
        }

        /**
         * Listen to `change` event on a model or emitter object and call `onChange` lifecycle method.
         * The listener will be removed when the component is destroyed.
         * By default the component will be subscribed to `this.model` and `this.state`.
         * @param {Rasti.Model} model A model or emitter object to listen to changes.
         * @return {Rasti.Component} The component instance.
         */
        subscribe(model) {
            // Check if model has `on` method.
            if (!model.on) return;
            // Store bound onChange method.
            const onChange = this.onChange.bind(this);
            // Listen to model changes and store unbind function.
            const off = model.on('change', onChange);
            // Add unbind function to destroy queue.
            // So the component stops listening to model changes when destroyed.
            this.destroyQueue.push(
                // Rasti `on` method returns an unbind function.
                // But other libraries may return the object itself.
                typeof off === 'function' ? off : () => model.off('change', onChange)
            );

            return this;
        }

        /**
         * Tell if `Component` is a container.
         * In which case, it will not have an element by itself.
         * It will render a single expression which is expected to return a single component as child.
         * `this.el` will be a reference to that child component's element.
         * @return {boolean}
         * @private
         */
        isContainer() {
            return !!(!this.tag && this.template);
        }

        /**
         * Override super method. We don't want to ensure an element on instantiation.
         * We will provide it later.
         * @private
         */
        ensureElement() {
            // If el is provided, delegate events.
            if (this.el) {
                // If "this.el" is a function, call it to get the element.
                this.el = getResult(this.el, this);
                this.delegateEvents();
            }
        }

        /**
         * Locate the root element of the `Component` within a specified parent node.
         * This is achieved by searching for the element using the unique data attribute assigned to the `Component`.
         * @param {Node} parent - The parent node to search within.
         * @return {Node} The root element of the component, or `null` if not found.
         * @private
         */
        findElement(parent) {
            return (parent || document).querySelector(`[${Component.DATA_ATTRIBUTE_UID}="${this.uid}"]`);
        }

        /**
         * Retrieve the attributes to be applied to the element.
         * This includes attributes to be added, removed, and their HTML representation.
         * @return {object} An object containing the following properties:
         * @property {object} add - Attributes to be added to the element, with their values.
         * @property {object} remove - Attributes to be removed from the element.
         * @property {string} html - A string representation of the attributes for use in HTML.
         * @private
         */
        getAttributes() {
            const add = {};
            const remove = {};
            const html = [];

            const attributes = { [Component.DATA_ATTRIBUTE_UID] : this.uid };

            if (this.attributes) Object.assign(attributes, getResult(this.attributes, this));
            // Store previous attributes.
            const previousAttributes = this.previousAttributes || {};
            this.previousAttributes = attributes;

            Object.keys(attributes).forEach(key => {
                let value = attributes[key];
                // Transform bool attribute values
                if (value === false) {
                    remove[key] = true;
                } else if (value === true) {
                    add[key] = '';
                    html.push(key);
                } else {
                    if (value === null || typeof value === 'undefined') value = '';

                    add[key] = value;
                    html.push(`${Component.sanitize(key)}="${Component.sanitize(value)}"`);
                }
            });
            // Remove attributes that were in previousAttributes but not in current attributes.
            Object.keys(previousAttributes).forEach(key => {
                if (!(key in attributes)) {
                    remove[key] = true;
                }
            });

            return { add, remove, html : html.join(' ') };
        }

        /**
         * Used internally on the render process.
         * Attach the `Component` to the dom element providing `this.el`, delegate events, 
         * subscribe to model changes and call `onRender` lifecycle method with `Component.RENDER_TYPE_HYDRATE` as argument.
         * @param parent {node} The parent node.
         * @return {Rasti.Component} The component instance.
         * @private
         */
        hydrate(parent) {
            // Listen to model changes and call onChange.
            if (this.model) this.subscribe(this.model);
            // Listen to state changes and call onChange.
            if (this.state) this.subscribe(this.state);

            if (this.isContainer()) {
                this.children[0].hydrate(parent);
                this.el = this.children[0].el;
            } else {
                this.el = this.findElement(parent);
                this.delegateEvents();
                this.children.forEach(child => child.hydrate(this.el));
            }
            // Call `onRender` lifecycle method.
            this.onRender.call(this, Component.RENDER_TYPE_HYDRATE);
            // Return `this` for chaining.
            return this;
        }

        /**
         * Used internally on the render process.
         * Reuse a `Component` that has `key` when its parent is rendered.
         * Call `onRender` lifecycle method with `Component.RENDER_TYPE_RECYCLE` as argument.
         * @param parent {node} The parent node.
         * @return {Rasti.Component} The component instance.
         * @private
         */
        recycle(parent) {
            // If component is a container, call recycle on its child.
            if (this.isContainer()) {
                this.children[0].recycle(parent);
            } else {
                // Find placeholder element to be replaced. It has same data attribute as this component.
                const toBeReplaced = this.findElement(parent);
                // Replace it with this.el.
                toBeReplaced.replaceWith(this.el);
            }
            // Call `onRender` lifecycle method.
            this.onRender.call(this, Component.RENDER_TYPE_RECYCLE);
            // Return `this` for chaining.
            return this;
        }

        /**
         * Destroy the `Component`.
         * Destroy children components if any, undelegate events, stop listening to events, call `onDestroy` lifecycle method.
         * @param {object} options Options object or any arguments passed to `destroy` method will be passed to `onDestroy` method.
         * @return {Rasti.View} Return `this` for chaining.
         */
        destroy() {
            // Call super destroy method.
            super.destroy.apply(this, arguments);
            // Set destroyed flag to prevent a last render after destroyed.
            this.destroyed = true;
            // Return `this` for chaining.
            return this;
        }

        /**
         * Lifecycle method. Called when the view is created, at the end of the `constructor`.
         * @param options {object} The view options.
         */
        onCreate() {}

        /**
         * Lifecycle method. Called when model emits `change` event.
         * By default calls `render` method.
         * This method can be extended with custom logic.
         * Maybe comparing new attributes with previous ones and calling
         * render when needed.
         * @param model {Rasti.Model} The model that emitted the event.
         * @param changed {object} Object containing keys and values that has changed.
         * @param [...args] {any} Any extra arguments passed to set method.
         */
        onChange() {
            this.render();
        }

        /**
         * Lifecycle method. Called after the component is rendered.
         * - When the component is rendered for the first time, this method is called with `Component.RENDER_TYPE_HYDRATE` as the argument.
         * - When the component is updated or re-rendered, this method is called with `Component.RENDER_TYPE_RENDER` as the argument.
         * - When the component is recycled (reused with the same key), this method is called with `Component.RENDER_TYPE_RECYCLE` as the argument.
         * @param {string} type - The render type. Possible values are: `Component.RENDER_TYPE_HYDRATE`, `Component.RENDER_TYPE_RENDER` and `Component.RENDER_TYPE_RECYCLE`.
         */
        onRender() {}

        /**
         * Lifecycle method. Called when the view is destroyed.
         * @param {object} options Options object or any arguments passed to `destroy` method.
         */
        onDestroy() {}

        /**
         * Tagged template helper method.
         * Used to create a partial template.  
         * It will return a one-dimensional array with strings and expressions.  
         * Components will be added as children by the parent component. Template strings literals 
         * will be marked as safe HTML to be rendered.
         * This method is bound to the component instance by default.
         * @param {TemplateStringsArray} strings - Template strings.
         * @param  {...any} expressions - Template expressions.
         * @return {Array} Array containing strings and expressions.
         * @example
         * import { Component } from 'rasti';
         * // Create a Title component.
         * const Title = Component.create`
         *     <h1>${self => self.renderChildren()}</h1>
         * `;
         * // Create Main component.
         * const Main = Component.create`
         *     <main>
         *         ${self => self.renderHeader()}
         *     </main>
         * `.extend({
         *     // Render header method.
         *     // Use `partial` to render an HTML template adding children components.
         *     renderHeader() {
         *         return this.partial`
         *             <header>
         *                 <${Title}>${({ model }) => model.title}</${Title}>
         *             </header>
         *         `;
         *     }
         * });
         */
        partial(strings, ...expressions) {
            return deepFlat(
                splitPlaceholders(
                    expandComponents(addPlaceholders(strings, expressions), expressions), expressions
                ).map(item => getExpressionResult(item, this))
            );
        }

        getRecyclePlaceholder() {
            if (this.isContainer()) return this.children[0].getRecyclePlaceholder();
            
            const tag = getResult(this.tag, this) || 'div';
            const attributes = `${Component.DATA_ATTRIBUTE_UID}="${this.uid}"`;

            return this.template || !selfClosingTags[tag] ?
                `<${tag} ${attributes}></${tag}>` :
                `<${tag} ${attributes} />`;
        }

        /*
         * Treat the whole view as a HTML string.
         */
        toString() {
            // Normally there won't be any children, but if there are, destroy them.
            this.destroyChildren();
            // Container.
            if (this.isContainer()) return this.template.call(this, this.addChild.bind(this));
            // Get tag name.
            const tag = getResult(this.tag, this) || 'div';
            // Get attributes.
            const attributes = this.getAttributes().html;
            // Replace expressions of inner template.
            const inner = this.template ? this.template.call(this, this.addChild.bind(this)) : '';
            // Generate outer template.
            return this.template || !selfClosingTags[tag] ?
                `<${tag} ${attributes}>${inner}</${tag}>` :
                `<${tag} ${attributes} />`;
        }

        /**
         * Render the `Component`.  
         * - If `this.el` is not present, the `Component` will be rendered as a string inside a `DocumentFragment` and hydrated, making `this.el` available. The `onRender` lifecycle method will be called with `Component.RENDER_TYPE_HYDRATE` as an argument.  
         * - If `this.el` is present, the method will update the attributes and inner HTML of the element, or recreate its child component in the case of a container. The `onRender` lifecycle method will be called with `Component.RENDER_TYPE_RENDER` as an argument.  
         * - When rendering child components, if the new children have the same key as the previous ones, they will be recycled. A recycled `Component` will call the `onRender` lifecycle method with `Component.RENDER_TYPE_RECYCLE` as an argument.  
         * - If the active element is inside the component, it will retain focus after the render.  
         * @return {Rasti.Component} The component instance.
         */
        render() {
            // Prevent a last re render if view is already destroyed.
            if (this.destroyed) return this;
            // If `this.el` is not present, render the view as a string and hydrate it.
            if (!this.el) {
                const fragment = this.createElement('template');
                fragment.innerHTML = this;
                this.hydrate(fragment.content);
                return this;
            }
            // Update attributes.
            if (!this.isContainer()) {
                const attributes = this.getAttributes();
                // Remove attributes.
                Object.keys(attributes.remove).forEach(key => {
                    this.el.removeAttribute(key);
                });
                // Add attributes.
                Object.keys(attributes.add).forEach(key => {
                    this.el.setAttribute(key, attributes.add[key]);
                });
            }
            // Check for `template` to see if view has innerHTML or a child component.
            if (this.template) {
                // Store active element.
                const activeElement = document.activeElement;

                const nextChildren = [];
                const recycledChildren = [];

                const previousChildren = this.children;
                this.children = [];
                // Replace expressions. Set html inside of `this.el`.
                const inner = this.template.call(this, component => {
                    let out = component;
                    // Check if child already exists.
                    const found = component.key && previousChildren.find(
                        previousChild => previousChild.key === component.key
                    );

                    if (found) {
                        // If child already exists, replace it html by its root element.
                        out = found.getRecyclePlaceholder();
                        // Add child to recycled children.
                        recycledChildren.push(found);
                        // Destroy new child component. Use recycled one instead.
                        component.destroy();
                    } else {
                        // Not found. Add new child component.
                        nextChildren.push(component);
                    }
                    // Component html.
                    return out;
                });

                if (this.isContainer()) {
                    if (nextChildren[0]) {
                        const fragment = this.createElement('template');
                        fragment.innerHTML = inner;
                        // Add new child to dom fragment and hydrate it.
                        this.addChild(nextChildren[0]).hydrate(fragment.content);
                        // Get next child element.
                        const nextEl = fragment.content.children[0];
                        // Replace `this.el` with nextEl.
                        this.el.replaceWith(nextEl);
                        // Set `this.el` to nextEl.
                        this.el = nextEl;
                    } else if (recycledChildren[0]) {
                        this.addChild(recycledChildren[0]);
                    } else {
                        throw new Error('Container component must have a child component');
                    }
                } else {
                    this.el.innerHTML = inner;
                    // Add new children. Hydrate them.
                    nextChildren.forEach(nextChild => {
                        this.addChild(nextChild).hydrate(this.el);
                    });
                    // Replace children root elements with recycled components.
                    recycledChildren.forEach(recycledChild => {
                        this.addChild(recycledChild).recycle(this.el);
                    });
                }
                // Destroy unused children.
                previousChildren.forEach(previousChild => {
                    const found = recycledChildren.indexOf(previousChild) > -1;
                    if (!found) previousChild.destroy();
                });
                // Restore focus.
                if (this.el.contains(activeElement)) {
                    activeElement.focus();
                }
            }
            // Call onRender lifecycle method.
            this.onRender.call(this, Component.RENDER_TYPE_RENDER);
            // Return this for chaining.
            return this;
        }

        /**
         * Mark a string as safe HTML to be rendered.  
         * Normally you don't need to use this method, as Rasti will automatically mark string literals 
         * as safe HTML when the component is {@link #module_component_create created} and when 
         * using the {@link #module_component__partial Component.partial} method.  
         * Be sure that the string is safe to be rendered, as it will be inserted into the DOM without any sanitization.
         * @static
         * @param {string} value 
         * @return {Rasti.SafeHTML} A safe HTML object.
         */
        static markAsSafeHTML(value) {
            return new SafeHTML(value);
        }

        /**
         * Helper method used to extend a `Component`, creating a subclass.
         * @static
         * @param {object} object Object containing methods to be added to the new `Component` subclass. Also can be a function that receives the parent prototype and returns an object.
         */
        static extend(object) {
            const Current = this;

            class Component extends Current {}

            Object.assign(
                Component.prototype,
                typeof object === 'function' ? object(Current.prototype) : object
            );

            return Component;
        }

        /**
         * Mount the component into the dom.
         * It instantiate the Component view using options, 
         * appends its element into the DOM (if `el` is provided).
         * And returns the view instance.
         * @static
         * @param {object} options The view options.
         * @param {node} el Dom element to append the view element.
         * @param {boolean} hydrate If true, the view will hydrate existing DOM.
         * @return {Rasti.Component}
         */
        static mount(options, el, hydrate) {
            // Instantiate view.
            const view = new this(options);
            // If `el` is passed, mount component.
            if (el) {
                if (hydrate) {
                    // Hydrate existing DOM.
                    view.toString();
                    view.hydrate(el);
                } else {
                    // Append element to the DOM.
                    el.appendChild(view.render().el);
                }
            }
            // Return view instance.
            return view;
        }

        /**
         * Takes a tagged template string or a function that returns another component, and returns a new `Component` class.
         * - The template outer tag and attributes will be used to create the view's root element.
         * - The template inner HTML will be used as the view's template.
         *   ```javascript
         *   const Button = Component.create`<button class="button">Click me</button>`;
         *   ```
         * - Template interpolations that are functions will be evaluated during the render process, receiving the view instance as an argument and being bound to it. If the function returns `null`, `undefined`, `false`, or an empty string, the interpolation won't render any content.
         *   ```javascript
         *   const Button = Component.create`
         *       <button class="${({ options }) => options.className}">
         *           ${({ options }) => options.renderChildren()}
         *       </button>
         *   `;
         *   ```
         * - Event handlers should be passed, at the root element as camelized attributes, in the format `onEventName=${{'selector' : listener }}`. They will be transformed to an event object and delegated to the root element. See {@link #module_view__delegateevents View.delegateEvents}. 
         * - Boolean attributes should be passed in the format `attribute="${() => true}"`. `false` attributes won't be rendered. `true` attributes will be rendered without a value.
         *   ```javascript
         *   const Input = Component.create`
         *       <input type="text" disabled=${({ options }) => options.disabled} />
         *   `;
         *   ```
         * - If the interpolated function returns a component instance, it will be added as a child component.
         * - If the interpolated function returns an array, each item will be evaluated as above.
         *   ```javascript
         *   // Create a button component.
         *   const Button = Component.create`
         *       <button class="button">
         *           ${({ options }) => options.renderChildren()}
         *       </button>
         *   `;
         *   // Create a navigation component. Add buttons as children. Iterate over items.
         *   const Navigation = Component.create`
         *       <nav>
         *           ${({ options }) => options.items.map(
         *               item => Button.mount({ renderChildren: () => item.label })
         *           )}
         *       </nav>
         *   `;
         *   // Create a header component. Add navigation as a child.
         *   const Header = Component.create`
         *       <header>
         *           ${({ options }) => Navigation.mount({ items : options.items})}
         *       </header>
         *   `;
         *   ```
         * - Child components can be added using a component tag.
         *   ```javascript
         *   // Create a button component.
         *   const Button = Component.create`
         *       <button class="button">
         *            ${({ options }) => options.renderChildren()}
         *       </button>
         *   `;
         *   // Create a navigation component. Add buttons as children. Iterate over items.
         *   const Navigation = Component.create`
         *       <nav>
         *           ${({ options, partial }) => options.items.map(
         *               item => partial`<${Button}>${item.label}</${Button}>`
         *           )}
         *       </nav>
         *   `;
         *   // Create a header component. Add navigation as a child.
         *   const Header = Component.create`
         *       <header>
         *           <${Navigation} items="${({ options }) => options.items}" />
         *       </header>
         *   `;
         *   ```
         * - If the tagged template contains only one expression that mounts a component, or the tags are references to a component, the component will be considered a <b>container</b>. It will render a single component as a child. `this.el` will be a reference to that child component's element.
         *   ```javascript
         *   // Create a button component.
         *   const Button = Component.create`
         *       <button class="${({ options }) => options.className}">
         *           ${self => self.renderChildren()}
         *       </button>
         *   `;
         *   // Create a container that renders a Button component.
         *   const ButtonOk = Component.create`
         *       <${Button} className="ok">Ok</${Button}>
         *   `;
         *   // Create a container that renders a Button component, using a function.
         *   const ButtonCancel = Component.create(() => Button.mount({
         *       className: 'cancel',
         *       renderChildren: () => 'Cancel'
         *   }));
         *   ```
         * @static
         * @param {string|function} strings - HTML template for the component or a function that mounts a sub component.
         * @param {...*} expressions - The expressions to be interpolated within the template.
         * @return {Rasti.Component} The newly created component class.
         */
        static create(strings, ...expressions) {
            const PH = Component.PLACEHOLDER_EXPRESSION('(\\d+)');
            // Containers can be created using create as a functions instead of a tagged template.
            if (typeof strings === 'function') {
                expressions = [strings];
                strings = ['', ''];
            }

            let tag, attributes, events, template;
            // Create output string for main template. Add placeholders for new lines.
            const main = expandComponents(addPlaceholders(strings, expressions), expressions);

            let match = main.match(
                new RegExp(`^\\s*<([a-z]+[1-6]?|${PH})([^>]*)>([\\s\\S]*?)</(\\1|${PH})>\\s*$|^\\s*<([a-z]+[1-6]?|${PH})([^>]*)/>\\s*$`)
            );

            if (match) {
                // It's a component with tag.
                const { tag : tagExpression, attributes : attributesAndEvents, inner, close } = parseMatch(match, expressions);
                // Get tag, attributes.
                tag = function() {
                    return Component.sanitize(getExpressionResult(tagExpression, this));
                };
                // Get attributes.
                attributes = function() {
                    return expandAttributes(attributesAndEvents, value => getExpressionResult(value, this)).attributes;
                };
                // Get events.
                events = function() {
                    const onlyEvents = expandAttributes(attributesAndEvents, value => getExpressionResult(value, this)).events;

                    return Object.keys(onlyEvents).reduce((out, key) => {
                        const typeListeners = getExpressionResult(onlyEvents[key], this);

                        Object.keys(typeListeners).forEach(selector => {
                            out[`${key}${selector === '&' ? '' : ` ${selector}`}`] = typeListeners[selector];
                        });

                        return out;
                    }, {});
                };
                // If there is a closing tag, get template.
                if (close) {
                    const list = inner ? splitPlaceholders(inner, expressions) : [];
                    template = function(addChild) {
                        return deepFlat(list.map(item => getExpressionResult(item, this))).map(item => {
                            if (typeof item !== 'undefined' && item !== null && item !== false &&  item !== true) {
                                if (item instanceof SafeHTML) return item;
                                if (item instanceof Component) return addChild(item);
                                return Component.sanitize(item);
                            }
                            return '';
                        }).join('');
                    };
                }
            } else {
                // It's a container.
                match = main.match(new RegExp(`^\\s*${PH}\\s*$`));

                if (match) {
                    // If there is only one expression and no tag, is a container.
                    template = function(addChild) {
                        // Replace expressions.
                        return addChild(getExpressionResult(expressions[match[1]], this)).toString();
                    };
                } else {
                    throw new SyntaxError('Invalid component');
                }
            }

            const Current = this;
            // Create subclass for this component.
            return Current.extend({
                // Set root element tag.
                tag,
                // Set attributes.
                attributes,
                // Set events.
                events,
                // Set template.
                template
            });
        }
    }

    Component.PLACEHOLDER_EXPRESSION = (idx) => `__RASTI_{${idx}}__`;
    Component.DATA_ATTRIBUTE_UID = 'data-rasti-uid';
    Component.RENDER_TYPE_HYDRATE = 'hydrate';
    Component.RENDER_TYPE_RECYCLE = 'recycle';
    Component.RENDER_TYPE_RENDER = 'render';

    exports.Component = Component;
    exports.Emitter = Emitter;
    exports.Model = Model;
    exports.View = View;

}));
